!-------------------------------------------------------------------------
! f90gl Version 1.0 September 1996
! 
! William F. Mitchell
! Applied and Computational Mathematics Division
! National Institute of Standards and Technology
! Gaithersburg, MD 20899
! william.mitchell@nist.gov
! 
! This fortran interface for Mesa software and documentation have been produced
! as part of work done by the U.S. Government, and are not subject to copyright
! in the United States.
! 
! The mention of specific products, trademarks, or brand names in the
! documentation is for purposes of identification only.  Such mention is not to
! be interpreted in any way as an endorsement or certification of such products
! or brands by the National Institute of Standards and Technology.
! All trademarks mentioned herein belong to their respective owners.
!-------------------------------------------------------------------------

 ! Draw a gear wheel.  You'll probably want to call this function when
 ! building a display list since we do a lot of trig here.
 !
 ! Input:  inner_radius - radius of hole at center
 !         outer_radius - radius at center of teeth
 !         width - width of gear
 !         teeth - number of teeth
 !         tooth_depth - depth of tooth

subroutine gear(inner_radius,outer_radius,width,teeth,tooth_depth)
use f90gl_kinds
use f90gl
real(f90glfloat) inner_radius, outer_radius, width, tooth_depth
integer(f90glint) teeth
real, parameter :: M_PI = 3.14159265

   integer(f90GLint) i
   real(f90glfloat) r0, r1, r2
   real(f90glfloat) angle, da
   real(f90glfloat) u, v, len

   r0 = inner_radius
   r1 = outer_radius - tooth_depth/2.0
   r2 = outer_radius + tooth_depth/2.0

   da = 2.0*M_PI / teeth / 4.0

   call f90glShadeModel( GL_FLAT )

   call f90glNormal3f( 0.0_f90glfloat, 0.0_f90glfloat, 1.0_f90glfloat )

   ! draw front face
   call f90glBegin( GL_QUAD_STRIP )
   do i = 0, teeth
      angle = i * 2.0*M_PI / teeth
      call f90glVertex3f( r0*cos(angle), r0*sin(angle), width*0.5 )
      call f90glVertex3f( r1*cos(angle), r1*sin(angle), width*0.5 )
      call f90glVertex3f( r0*cos(angle), r0*sin(angle), width*0.5 )
      call f90glVertex3f( r1*cos(angle+3*da), r1*sin(angle+3*da), width*0.5 )
   end do
   call f90glEnd()

   ! draw front sides of teeth
   call f90glBegin( GL_QUADS )
   da = 2.0*M_PI / teeth / 4.0;
   do i = 0, teeth-1
      angle = i * 2.0*M_PI / teeth

      call f90glVertex3f( r1*cos(angle),      r1*sin(angle),      width*0.5 )
      call f90glVertex3f( r2*cos(angle+da),   r2*sin(angle+da),   width*0.5 )
      call f90glVertex3f( r2*cos(angle+2*da), r2*sin(angle+2*da), width*0.5 )
      call f90glVertex3f( r1*cos(angle+3*da), r1*sin(angle+3*da), width*0.5 )
   end do
   call f90glEnd()


   call f90glNormal3f( 0.0_f90glfloat, 0.0_f90glfloat, -1.0_f90glfloat )

   ! draw back face
   call f90glBegin( GL_QUAD_STRIP )
   do i = 0, teeth
      angle = i * 2.0*M_PI / teeth
      call f90glVertex3f( r1*cos(angle), r1*sin(angle), -width*0.5 )
      call f90glVertex3f( r0*cos(angle), r0*sin(angle), -width*0.5 )
      call f90glVertex3f( r1*cos(angle+3*da), r1*sin(angle+3*da), -width*0.5 )
      call f90glVertex3f( r0*cos(angle), r0*sin(angle), -width*0.5 )
   end do
   call f90glEnd()

   ! draw back sides of teeth
   call f90glBegin( GL_QUADS )
   da = 2.0*M_PI / teeth / 4.0
   do i = 0, teeth-1
      angle = i * 2.0*M_PI / teeth

      call f90glVertex3f( r1*cos(angle+3*da), r1*sin(angle+3*da), -width*0.5 )
      call f90glVertex3f( r2*cos(angle+2*da), r2*sin(angle+2*da), -width*0.5 )
      call f90glVertex3f( r2*cos(angle+da),   r2*sin(angle+da),   -width*0.5 )
      call f90glVertex3f( r1*cos(angle),      r1*sin(angle),      -width*0.5 )
   end do
   call f90glEnd()


   ! draw outward faces of teeth
   call f90glBegin( GL_QUAD_STRIP )
   do i = 0, teeth-1
      angle = i * 2.0*M_PI / teeth

      call f90glVertex3f( r1*cos(angle),      r1*sin(angle),       width*0.5 )
      call f90glVertex3f( r1*cos(angle),      r1*sin(angle),      -width*0.5 )
      u = r2*cos(angle+da) - r1*cos(angle)
      v = r2*sin(angle+da) - r1*sin(angle)
      len = sqrt( u*u + v*v )
      u = u/len
      v = v/len
      call f90glNormal3f( v, -u, 0.0_f90glfloat )
      call f90glVertex3f( r2*cos(angle+da),   r2*sin(angle+da),    width*0.5 )
      call f90glVertex3f( r2*cos(angle+da),   r2*sin(angle+da),   -width*0.5 )
      call f90glNormal3f( cos(angle), sin(angle), 0.0_f90glfloat )
      call f90glVertex3f( r2*cos(angle+2*da), r2*sin(angle+2*da),  width*0.5 )
      call f90glVertex3f( r2*cos(angle+2*da), r2*sin(angle+2*da), -width*0.5 )
      u = r1*cos(angle+3*da) - r2*cos(angle+2*da)
      v = r1*sin(angle+3*da) - r2*sin(angle+2*da)
      call f90glNormal3f( v, -u, 0.0_f90glfloat )
      call f90glVertex3f( r1*cos(angle+3*da), r1*sin(angle+3*da),  width*0.5 )
      call f90glVertex3f( r1*cos(angle+3*da), r1*sin(angle+3*da), -width*0.5 )
      call f90glNormal3f( cos(angle), sin(angle), 0.0_f90glfloat )
   end do

   call f90glVertex3f( r1*cos(0.), r1*sin(0.), width*0.5 )
   call f90glVertex3f( r1*cos(0.), r1*sin(0.), -width*0.5 )

   call f90glEnd()


   call f90glShadeModel( GL_SMOOTH )

   ! draw inside radius cylinder
   call f90glBegin( GL_QUAD_STRIP )
   do i = 0, teeth
      angle = i * 2.0*M_PI / teeth
      call f90glNormal3f( -cos(angle), -sin(angle), 0.0_f90glfloat )
      call f90glVertex3f( r0*cos(angle), r0*sin(angle), -width*0.5 )
      call f90glVertex3f( r0*cos(angle), r0*sin(angle), width*0.5 )
   end do
   call f90glEnd()
   return
      
end subroutine


module gear_globals
use f90gl_kinds

real(f90glfloat), save :: view_rotx=20.0, view_roty=30.0, view_rotz=0.0
integer(f90glint), save :: gear1, gear2, gear3
real(f90glfloat), save :: angle = 0.0;

integer(f90gluint), save :: limit
integer(f90gluint), save :: count = 1

end module

subroutine draw
use gear_globals
use f90gl
use f90gltk

   call f90glClear( ior(GL_COLOR_BUFFER_BIT, GL_DEPTH_BUFFER_BIT) )

! taking a chance that f90glfloat is the default kind of real

   call f90glPushMatrix()
   call f90glRotatef( view_rotx, 1.0, 0.0, 0.0 )
   call f90glRotatef( view_roty, 0.0, 1.0, 0.0 )
   call f90glRotatef( view_rotz, 0.0, 0.0, 1.0 )

   call f90glPushMatrix()
   call f90glTranslatef( -3.0, -2.0, 0.0 )
   call f90glRotatef( angle, 0.0, 0.0, 1.0 )
   call f90glCallList(gear1)
   call f90glPopMatrix()

   call f90glPushMatrix()
   call f90glTranslatef( 3.1, -2.0, 0.0 )
   call f90glRotatef( -2.0*angle-9.0, 0.0, 0.0, 1.0 )
   call f90glCallList(gear2)
   call f90glPopMatrix()

   call f90glPushMatrix()
   call f90glTranslatef( -3.1, 4.2, 0.0 )
   call f90glRotatef( -2.0*angle-25.0, 0.0, 0.0, 1.0 )
   call f90glCallList(gear3)
   call f90glPopMatrix()

   call f90glPopMatrix()

   call f90tkSwapBuffers()

   count = count+1
   if (count==limit) call f90tkQuit()
   return
end subroutine



subroutine idle
use gear_globals
   angle = angle + 2.0
   call draw()
   return
end subroutine



! change view angle, exit upon ESC 
function key(k, mask)
use gear_globals
use f90gl
use f90gltk
integer(f90glenum) key
integer(f90glcint) k
integer(f90glenum) mask

   key = GL_FALSE
   select case(k)
      case (TK_UP)
         view_rotx = view_rotx + 5.0
         key = GL_TRUE
      case (TK_DOWN)
         view_rotx = view_rotx - 5.0
         key = GL_TRUE
      case (TK_LEFT)
         view_roty = view_roty + 5.0
         key = GL_TRUE
      case (TK_RIGHT)
         view_roty = view_roty - 5.0
         key = GL_TRUE
      case (TK_lc_z)
         view_rotz = view_rotz + 5.0
         key = GL_TRUE
      case (TK_Z)
         view_rotz = view_rotz - 5.0
         key = GL_TRUE
      case (TK_ESCAPE)
         call f90tkQuit()
   end select
   return
end function



! new window size or exposure
subroutine reshape( width, height )
use f90gl_kinds
use f90gl
integer(f90glcint) width, height
real(f90gldouble) h
integer(f90glsizei) swidth, sheight

   h = real(height,kind=f90gldouble) / real(width,kind=f90gldouble)
   swidth = width; sheight = height

   call f90glViewport(0_f90glint, 0_f90glint, swidth, sheight)
   call f90glMatrixMode(GL_PROJECTION)
   call f90glLoadIdentity()
   call f90glFrustum( -1.0_f90gldouble, 1.0_f90gldouble, -h, h, &
                      5.0_f90gldouble, 60.0_f90gldouble )
   call f90glMatrixMode(GL_MODELVIEW)
   call f90glLoadIdentity()
   call f90glTranslatef( 0.0_f90glfloat, 0.0_f90glfloat, -40.0_f90glfloat )
   call f90glClear( ior(GL_COLOR_BUFFER_BIT, GL_DEPTH_BUFFER_BIT) )
   return

end subroutine


subroutine init
use gear_globals
use f90gl

   real(f90glfloat), save :: pos(4) = (/5.0, 5.0, 10.0, 0.0 /)
   real(f90glfloat), save :: red(4) = (/0.8, 0.1, 0.0, 1.0 /)
   real(f90glfloat), save :: green(4) = (/0.0, 0.8, 0.2, 1.0 /)
   real(f90glfloat), save :: blue(4) = (/0.2, 0.2, 1.0, 1.0 /)

   call f90glLightfv( GL_LIGHT0, GL_POSITION, pos )
   call f90glEnable( GL_CULL_FACE )
   call f90glEnable( GL_LIGHTING )
   call f90glEnable( GL_LIGHT0 )
   call f90glEnable( GL_DEPTH_TEST )

   ! make the gears
   gear1 = f90glGenLists(1)
   call f90glNewList(gear1, GL_COMPILE)
   call f90glMaterialfv( GL_FRONT, GL_AMBIENT_AND_DIFFUSE, red )
! taking a chance that f90glfloat and f90glint are default kind
   call gear( 1.0, 4.0, 1.0, 20, 0.7 )
   call f90glEndList()

   gear2 = f90glGenLists(1)
   call f90glNewList(gear2, GL_COMPILE)
   call f90glMaterialfv( GL_FRONT, GL_AMBIENT_AND_DIFFUSE, green )
   call gear( 0.5, 2.0, 2.0, 10, 0.7 )
   call f90glEndList()

   gear3 = f90glGenLists(1)
   call f90glNewList(gear3, GL_COMPILE)
   call f90glMaterialfv( GL_FRONT, GL_AMBIENT_AND_DIFFUSE, blue )
   call gear( 1.3, 2.0, 0.5, 10, 0.7 )
   call f90glEndList()

   call f90glEnable( GL_NORMALIZE )
   return
end subroutine


program gears
use gear_globals
use f90gl
use f90gltk
external reshape, idle, draw
integer(f90glenum), external :: key

   limit = 0
   call f90tkInitPosition(0_f90glcint, 0_f90glcint, 300_f90glcint, 300_f90glcint)
   call f90tkInitDisplayMode( ior(TK_RGB,ior(TK_DEPTH,ior(TK_DOUBLE,TK_DIRECT))) )

   if (f90tkInitWindow("Gears") == GL_FALSE) call f90tkQuit()

   call init()

   call f90tkExposeFunc( reshape )
   call f90tkReshapeFunc( reshape )
   call f90tkKeyDownFunc( key )
   call f90tkIdleFunc( idle )
   call f90tkDisplayFunc( draw )
   call f90tkExec()

end program
