!-------------------------------------------------------------------------
! f90gl Version 1.0 September 1996
! 
! William F. Mitchell
! Applied and Computational Mathematics Division
! National Institute of Standards and Technology
! Gaithersburg, MD 20899
! william.mitchell@nist.gov
! 
! This fortran interface for Mesa software and documentation have been produced
! as part of work done by the U.S. Government, and are not subject to copyright
! in the United States.
! 
! The mention of specific products, trademarks, or brand names in the
! documentation is for purposes of identification only.  Such mention is not to
! be interpreted in any way as an endorsement or certification of such products
! or brands by the National Institute of Standards and Technology.
! All trademarks mentioned herein belong to their respective owners.
!-------------------------------------------------------------------------

 ! Demo of a reflective, texture-mapped surface with OpenGL.
 ! Brian Paul  (brianp@ssec.wisc.edu)  August 14, 1995
 ! Fortran 90 by Bill Mitchell (mitchell@cam.nist.gov)  August 26, 1996
 !
 ! Hardware texture mapping is highly recommended!
 !
 ! The basic steps are:
 !    1. Render the reflective object (a polygon) from the normal viewpoint,
 !       setting the stencil planes = 1.
 !    2. Render the scene from a special viewpoint:  the viewpoint which
 !       is on the opposite side of the reflective plane.  Only draw where
 !       stencil = 1.  This draws the objects in the reflective surface.
 !    3. Render the scene from the original viewpoint.  This draws the
 !       objects in the normal fashion.  Use blending when drawing
 !       the reflective, textured surface.
 !
 ! This is a very crude demo.  It could be much better.
 
 ! Dirk Reiners (reiners@igd.fhg.de) made some modifications to the C code.

!#define USE_ZBUFFER

! OK, without hardware support this is overkill.
!#define USE_TEXTURE


module reflect_mod
use f90gl
use f90glu
use f90gltk

real(f90glfloat), parameter :: DEG2RAD = 3.14159/180.0
character(len=19) :: TABLE_TEXTURE = "samples/1.rgb"
integer(f90glint), parameter :: MAX_OBJECTS = 2

integer(f90glint), save :: table_list, objects_list(0:MAX_OBJECTS-1)
real(f90glfloat), save :: xrot, yrot, spin

contains

subroutine make_table

   real(f90glfloat), save :: table_mat(4) = (/ 1.0, 1.0, 1.0, 0.6 /)
   real(f90glfloat), save :: white(4) = (/ 1.0, 1.0, 1.0, 1.0 /)
   real(f90glfloat), save :: gray(4) = (/ 0.4, 0.4, 0.4, 1.0 /)
   type (f90tk_rgbimagerec), pointer :: image
   integer(f90glint) err

   table_list = f90glGenLists(1)
   call f90glNewList( table_list, GL_COMPILE )

   ! load table's texture
   call f90glMaterialfv( GL_FRONT, GL_AMBIENT_AND_DIFFUSE, table_mat )
!   call f90glMaterialfv( GL_FRONT, GL_EMISSION, gray )
   call f90glMaterialfv( GL_FRONT, GL_DIFFUSE, table_mat )
   call f90glMaterialfv( GL_FRONT, GL_AMBIENT, gray )
   
!#ifdef USE_TEXTURE
   call f90glEnable( GL_TEXTURE_2D )

   image => f90tkRGBImageLoad( TABLE_TEXTURE )
   err = f90gluBuild2DMipmaps(GL_TEXTURE_2D, 3, image%sizeX, image%sizeY, &
                     GL_RGB, GL_UNSIGNED_BYTE, image%data)

   call f90glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT )
   call f90glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT )
!#endif

   ! draw textured square for the table
   ! take a chance that f90glfloat is the default kind
   call f90glPushMatrix()
   call f90glScalef( 4.0, 4.0, 4.0 )
   call f90glBegin( GL_POLYGON )
   call f90glNormal3f( 0.0, 1.0, 0.0 )
   call f90glTexCoord2f( 0.0, 0.0 ); call  f90glVertex3f( -1.0, 0.0,  1.0 )
   call f90glTexCoord2f( 1.0, 0.0 ); call  f90glVertex3f(  1.0, 0.0,  1.0 )
   call f90glTexCoord2f( 1.0, 1.0 ); call  f90glVertex3f(  1.0, 0.0, -1.0 )
   call f90glTexCoord2f( 0.0, 1.0 ); call  f90glVertex3f( -1.0, 0.0, -1.0 )
   call f90glEnd()
   call f90glPopMatrix()

   call f90glDisable( GL_TEXTURE_2D )

   call f90glEndList()
   return
end subroutine


subroutine make_objects
   type (f90GLUquadricObj), pointer :: q

   real(f90glfloat), save :: cyan(4) = (/ 0.0, 1.0, 1.0, 1.0 /)
   real(f90glfloat), save :: green(4) = (/ 0.2, 1.0, 0.2, 1.0 /)
   real(f90glfloat), save :: black(4) = (/ 0.0, 0.0, 0.0, 0.0 /)

   q => f90gluNewQuadric()
   call f90gluQuadricDrawStyle( q, GLU_FILL )
   call f90gluQuadricNormals( q, GLU_SMOOTH )

   objects_list(0) = f90glGenLists(1)
   call f90glNewList( objects_list(0), GL_COMPILE )
   call f90glMaterialfv( GL_FRONT, GL_AMBIENT_AND_DIFFUSE, cyan )
   call f90glMaterialfv( GL_FRONT, GL_EMISSION, black )
   call f90gluCylinder( q, 0.5_f90gldouble, 0.5_f90gldouble,  1.0_f90gldouble, &
                        15, 10 )
   call f90glEndList()

   objects_list(1) = f90glGenLists(1)
   call f90glNewList( objects_list(1), GL_COMPILE )
   call f90glMaterialfv( GL_FRONT, GL_AMBIENT_AND_DIFFUSE, green )
   call f90glMaterialfv( GL_FRONT, GL_EMISSION, black )
   call f90gluCylinder( q, 1.5_f90gldouble, 0.0_f90gldouble,  2.5_f90gldouble, &
                        15, 10 )
   call f90glEndList()
   return
end subroutine

subroutine init

   real(f90glfloat), save :: light_pos(4) = (/ 0.0, 0.0, 20.0, 1.0 /)

   call make_table()
   call make_objects()
   xrot = 30.0
   yrot = 50.0
   spin = 0.0

!#ifndef USE_ZBUFFER
!   call f90glEnable( GL_CULL_FACE )
!#endif

   call f90glShadeModel( GL_FLAT )
   
   call f90glEnable( GL_LIGHT0 )
   call f90glEnable( GL_LIGHTING )

   call f90glLightfv( GL_LIGHT0, GL_POSITION, light_pos )
 
   call f90glClearColor( 0.5_f90glclampf, 0.5_f90glclampf, 0.5_f90glclampf, &
                         1.0_f90glclampf )

   call f90glEnable( GL_NORMALIZE )
   return
end subroutine

subroutine draw_objects( eyex, eyey, eyez )
real(f90gldouble) eyex, eyey, eyez

!#ifndef USE_ZBUFFER
!       if (eyex<0.5) then
!#endif
! taking a chance that f90glfloat is the default kind
           call f90glPushMatrix()
           call f90glTranslatef( 1.0, 1.5, 0.0 )
           call f90glRotatef( spin, 1.0, 0.5, 0.0 )
           call f90glRotatef( 0.5*spin, 0.0, 0.5, 1.0 )
           call f90glCallList( objects_list(0) )
           call f90glPopMatrix()

           call f90glPushMatrix()
           call f90glTranslatef( -1.0, 0.85+3.0*abs( cos(0.01*spin) ), 0.0 )
           call f90glRotatef( 0.5*spin, 0.0, 0.5, 1.0 )
           call f90glRotatef( spin, 1.0, 0.5, 0.0 )
           call f90glScalef( 0.5, 0.5, 0.5 )
           call f90glCallList( objects_list(1) )
           call f90glPopMatrix()
!#ifndef USE_ZBUFFER
!       else
!          call f90glPushMatrix()
!          call f90glTranslatef( -1.0, 0.85+3.0*abs( cos(0.01*spin) ), 0.0 )
!          call f90glRotatef( 0.5*spin, 0.0, 0.5, 1.0 )
!          call f90glRotatef( spin, 1.0, 0.5, 0.0 )
!          call f90glScalef( 0.5, 0.5, 0.5 )
!          call f90glCallList( objects_list(1) )
!          call f90glPopMatrix()
!
!          call f90glPushMatrix()
!          call f90glTranslatef( 1.0, 1.5, 0.0 )
!          call f90glRotatef( spin, 1.0, 0.5, 0.0 )
!          call f90glRotatef( 0.5*spin, 0.0, 0.5, 1.0 )
!          call f90glCallList( objects_list(0) )
!          call f90glPopMatrix()
!       endif
!#endif
   return
end subroutine


subroutine draw_table
   call f90glCallList( table_list )
   return
end subroutine



subroutine disp
   real(f90glfloat) :: dist = 20.0
   real(f90gldouble) :: eyex, eyey, eyez

   call f90glClear( ior(GL_COLOR_BUFFER_BIT, ior( GL_DEPTH_BUFFER_BIT,GL_STENCIL_BUFFER_BIT)) )


   eyex = dist * cos(yrot*DEG2RAD) * cos(xrot*DEG2RAD)
   eyez = dist * sin(yrot*DEG2RAD) * cos(xrot*DEG2RAD)
   eyey = dist * sin(xrot*DEG2RAD)

   ! view from top
   call f90glPushMatrix()
   call f90gluLookAt( eyex, eyey, eyez, &
                      0.0_f90gldouble, 0.0_f90gldouble, 0.0_f90gldouble,  &
                      0.0_f90gldouble, 1.0_f90gldouble, 0.0_f90gldouble )

   call draw_table()

   call f90glPopMatrix()

   call f90tkSwapBuffers()

   return
end subroutine

end module reflect_mod


subroutine display
use reflect_mod

   real (f90glfloat) :: dist = 20.0
   real (f90gldouble) eyex, eyey, eyez
   integer (f90gluint) all_ones
   data all_ones / z'ffffffff' /
   logical(f90glboolean) :: myGL_TRUE = .true., myGL_FALSE = .false.

   call f90glClear( ior(GL_COLOR_BUFFER_BIT, ior(GL_DEPTH_BUFFER_BIT, GL_STENCIL_BUFFER_BIT)) )


   eyex = dist * cos(yrot*DEG2RAD) * cos(xrot*DEG2RAD)
   eyez = dist * sin(yrot*DEG2RAD) * cos(xrot*DEG2RAD)
   eyey = dist * sin(xrot*DEG2RAD)

   ! view from top
   call f90glPushMatrix()
   call f90gluLookAt( eyex, eyey, eyez, &
                      0.0_f90gldouble, 0.0_f90gldouble, 0.0_f90gldouble, &
                      0.0_f90gldouble, 1.0_f90gldouble, 0.0_f90gldouble )


   ! draw table into stencil planes
   call f90glEnable( GL_STENCIL_TEST )
!#ifdef USE_ZBUFFER
   call f90glDisable( GL_DEPTH_TEST )
!#endif
   call f90glStencilFunc( GL_ALWAYS, 1, all_ones )
   call f90glStencilOp( GL_REPLACE, GL_REPLACE, GL_REPLACE )
   call f90glColorMask( myGL_FALSE, myGL_FALSE, myGL_FALSE, myGL_FALSE )
   call draw_table()
   call f90glColorMask( myGL_TRUE, myGL_TRUE, myGL_TRUE, myGL_TRUE )

!#ifdef USE_ZBUFFER
   call f90glEnable( GL_DEPTH_TEST )
!#endif


   ! render view from below (reflected viewport)
   ! only draw where stencil==1
   if (eyey>0.0) then
      call f90glPushMatrix()
 
      call f90glStencilFunc( GL_EQUAL, 1, all_ones )  ! draw if ==1
      call f90glStencilOp( GL_KEEP, GL_KEEP, GL_KEEP )
      call f90glScalef( 1.0, -1.0, 1.0 )
      call draw_objects(eyex, eyey, eyez)
      call f90glPopMatrix()
   endif

   call f90glDisable( GL_STENCIL_TEST )

   call f90glEnable( GL_BLEND )
   call f90glBlendFunc( GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA )

   call draw_table()
   call f90glDisable( GL_BLEND )

   ! view from top
   call f90glPushMatrix()

   call draw_objects(eyex, eyey, eyez)

   call f90glPopMatrix()

   call f90glPopMatrix()

   call f90tkSwapBuffers()

   return
end subroutine

subroutine reshape(w, h)
use reflect_mod
integer(f90glcint) w,h

   real(f90gldouble) aspect
   integer(f90glsizei) sw, sh

   aspect = real(w,kind=f90gldouble)/real(h,kind=f90gldouble)
   sw = w; sh=h

   call f90glViewport(0_f90glint, 0_f90glint, sw, sh)
   call f90glMatrixMode(GL_PROJECTION)
   call f90glLoadIdentity()
   call f90glFrustum( -aspect, aspect, -1.0_f90gldouble, 1.0_f90gldouble, &
                      4.0_f90gldouble, 300.0_f90gldouble )
   call f90glMatrixMode(GL_MODELVIEW)
   call f90glLoadIdentity()

   return
end subroutine

function key( k, mask )
use reflect_mod
integer(f90glenum) key
integer(f90glcint) k
integer(f90glenum) mask

   select case(k)
      case (27)
         call f90tkQuit

      case (TK_UP)
         xrot = xrot + 3.0
!#ifndef USE_ZBUFFER
!        if ( xrot > 180 ) xrot = 180
!#endif
      case (TK_DOWN)
         xrot = xrot - 3.0
!#ifndef USE_ZBUFFER
!        if ( xrot < 0 ) xrot = 0
!#endif
      case (TK_LEFT)
         yrot = yrot + 3.0
      case (TK_RIGHT)
         yrot = yrot - 3.0
   end select
   key = 0
   return
end function



subroutine idle
   use reflect_mod
   spin = spin + 2.0
   call display()
   return
end subroutine



program reflect
use reflect_mod
external reshape, display, idle
integer(f90glenum), external :: key
integer(f90glenum) dummy

! without USE_ZBUFFER, do not have TK_DEPTH
    call f90tkInitDisplayMode(ior(TK_DOUBLE, ior(TK_RGB, ior(TK_DEPTH, TK_STENCIL))) )
    call f90tkInitPosition( 0, 0, 400, 300 )
    dummy = f90tkInitWindow('reflect')
    call f90tkReshapeFunc(reshape)
    call f90tkDisplayFunc(display)
    call f90tkKeyDownFunc(key)
    call f90tkIdleFunc(idle)

    call init()

    call f90tkExec()
end program
