!-------------------------------------------------------------------------
! f90gl Version 1.0 September 1996
! 
! William F. Mitchell
! Applied and Computational Mathematics Division
! National Institute of Standards and Technology
! Gaithersburg, MD 20899
! william.mitchell@nist.gov
! 
! This fortran interface for Mesa software and documentation have been produced
! as part of work done by the U.S. Government, and are not subject to copyright
! in the United States.
! 
! The mention of specific products, trademarks, or brand names in the
! documentation is for purposes of identification only.  Such mention is not to
! be interpreted in any way as an endorsement or certification of such products
! or brands by the National Institute of Standards and Technology.
! All trademarks mentioned herein belong to their respective owners.
!-------------------------------------------------------------------------

module wave_mod
use f90gl
use f90glu
use f90gltk

! If the compiler complains about ambiguous generic interfaces, then remove
! f90glteximage2dbyte from the interface.  If it still complains, then
! remove the whole interface

interface f90glteximage2d
   module procedure f90glteximage2dshort, f90glteximage2dbyte
end interface

real(f90gldouble), parameter :: PI = 3.14159265358979323846
integer, parameter :: NULL=0

logical rgb, doubleBuffer, directRender

integer(f90GLint), target :: colorIndexes1(3), colorIndexes2(3)
integer(f90GLbitfield) :: clearMask = ior(GL_COLOR_BUFFER_BIT, GL_DEPTH_BUFFER_BIT)

logical :: smooth = .false., &
           lighting = .true., &
           depth = .true., &
           stepMode = .false., &
           spinMode = .true.
integer(f90GLint) :: contouring = 0

integer(f90GLint) widthX, widthY, checkerSize
real height

integer(f90GLint) :: frames, curFrame = 0, nextFrame = 0

type facet_t
    real color(3)
    real normal(3)
end type facet_t

type coord_t
    real vertex(3)
    real normal(3)
end type coord_t

type mesh
    integer(f90GLint) widthX, widthY
    integer(f90GLint) numFacets
    integer(f90GLint) numCoords
    integer(f90GLint) frames
    type(coord_t), pointer :: coords(:)
    type(facet_t), pointer :: facets(:)
end type mesh

type(mesh), target :: theMesh

integer(f90GLubyte) :: contourTexture1(16) = (/ &
    255, 255, 255, 255, &
    255, 255, 255, 255, &
    255, 255, 255, 255, &
    127, 127, 127, 127 &
/)
integer(f90GLubyte) :: contourTexture2(16) = (/ &
    255, 255, 255, 255, &
    255, 127, 127, 127, &
    255, 127, 127, 127, &
    255, 127, 127, 127 &
/)

contains

function GETCOORD(frame,x,y)
integer(f90glint) frame,x,y
type(coord_t), pointer :: GETCOORD
GETCOORD => theMesh%coords(frame*theMesh%numCoords + x + y*(theMesh%widthX+1) + 1)
return
end function GETCOORD

function GETFACET(frame,x,y)
integer(f90glint) frame,x,y
type(facet_t), pointer :: GETFACET
GETFACET => theMesh%facets(frame*theMesh%numFacets + x + y*theMesh%widthX + 1)
return
end function GETFACET

subroutine SetColorMap

    real, save, target :: green(3) = (/ 0.2, 1.0, 0.2 /)
    real, save, target :: red(3) = (/ 1.0, 0.2, 0.2 /)
    real, pointer :: color(:)
    real percent
    integer(f90GLint), pointer :: indexes(:)
    integer(f90glint) entries, i, j
    integer buf(4)

    entries = f90tkGetColorMapSize()

    colorIndexes1 = (/1, 1+int((entries-1)*0.3), int((entries-1)*0.5) /)
    colorIndexes2 = (/1+int((entries-1)*0.5), 1+int((entries-1)*0.8), entries-1 /)

    do i = 0, 1
        select case (i)
          case (0)
            color => green
            indexes => colorIndexes1
          case (1)
            color => red
            indexes => colorIndexes2
        end select

        do j = indexes(1), indexes(2)-1
            percent = 0.2 + 0.8 * (j - indexes(1)) / &
                      real(indexes(2) - indexes(1))
            call f90tkSetOneColor(j, percent*color(1), percent*color(2), &
                           percent*color(3))
        end do
        do j=indexes(2), indexes(3)-1
            percent = (j - indexes(2)) / real(indexes(3) - indexes(2))
            call f90tkSetOneColor(j, percent*(1-color(1))+color(1), &
                           percent*(1-color(2))+color(2), &
                           percent*(1-color(3))+color(3))
        end do
     end do
end subroutine

subroutine InitMesh

    type(coord_t), pointer :: coord, temppt
    type(facet_t), pointer :: facet
    real dp1(3), dp2(3)
    real, pointer :: pt1(:), pt2(:), pt3(:)
    real angle, d, x, y
    integer(f90GLint) numFacets, numCoords, frameNum, i, j
    integer erralloc

    theMesh%widthX = widthX
    theMesh%widthY = widthY
    theMesh%frames = frames

    numFacets = widthX * widthY
    numCoords = (widthX + 1) * (widthY + 1)

    theMesh%numCoords = numCoords
    theMesh%numFacets = numFacets

    allocate(theMesh%coords(frames*numCoords),stat=erralloc)
    allocate(theMesh%facets(frames*numFacets),stat=erralloc)
    if (erralloc /= 0) then
        print *,"Out of memory."
        call f90tkQuit()
    endif

    do frameNum = 0, frames-1
        do i = 0, widthX
            x = i / real(widthX)
            do j = 0, widthY
                y = j / real(widthY)

                d = sqrt(x*x+y*y)
                if (d == 0.0) d = 0.0001
                angle = 2 * PI * d + (2 * PI / frames * frameNum)

                coord => GETCOORD(frameNum, i, j)

                coord%vertex = (/ x-0.5, y-0.5, (height-height*d)*cos(angle) /)

                coord%normal = (/ -(height / d) * x * ((1 - d) * 2 * PI * &
                                     sin(angle) + cos(angle)), &
                                  -(height / d) * y * ((1 - d) * 2 * PI * &
                                     sin(angle) + cos(angle)), &
                                  -1._f90gldouble /)

                d = 1.0 / sqrt(coord%normal(1)*coord%normal(1)+ &
                               coord%normal(2)*coord%normal(2)+1)
                coord%normal = coord%normal * d
            end do
        end do
        do i = 0, widthX-1
            do j = 0, widthY-1
                facet => GETFACET(frameNum, i, j)
                if (ieor(mod((i/checkerSize),2),mod((j/checkerSize),2)) /= 0) then
                    if (rgb) then
                        facet%color = (/ 1.0, 0.2, 0.2 /)
                    else
                        facet%color = colorIndexes1
                    endif
                else
                    if (rgb) then
                        facet%color = (/ 0.2, 1.0, 0.2 /)
                    else
                        facet%color = colorIndexes2
                    endif
                endif
                temppt => GETCOORD(frameNum, i, j)
                pt1 => temppt%vertex
                temppt => GETCOORD(frameNum, i, j+1)
                pt2 => temppt%vertex
                temppt => GETCOORD(frameNum, i+1, j+1)
                pt3 => temppt%vertex

                dp1 = pt2 - pt1
                dp2 = pt3 - pt2

                facet%normal(1) = dp1(2) * dp2(3) - dp1(3) * dp2(2)
                facet%normal(2) = dp1(3) * dp2(1) - dp1(1) * dp2(3)
                facet%normal(3) = dp1(1) * dp2(2) - dp1(2) * dp2(1)

                d = 1.0 / sqrt(facet%normal(1)*facet%normal(1)+ &
                               facet%normal(2)*facet%normal(2)+ &
                               facet%normal(3)*facet%normal(3))

                facet%normal = facet%normal * d
            end do
        end do
    end do
end subroutine

subroutine InitMaterials

    real, save :: ambient(4) = (/0.1, 0.1, 0.1, 1.0/)
    real, save :: diffuse(4) = (/0.5, 1.0, 1.0, 1.0/)
    real, save :: position(4) = (/90.0, 90.0, 150.0, 0.0/)
    real, save :: front_mat_shininess(1) = (/60.0/)
    real, save :: front_mat_specular(4) = (/0.2, 0.2, 0.2, 1.0/)
    real, save :: front_mat_diffuse(4) = (/0.5, 0.28, 0.38, 1.0/)
    real, save :: back_mat_shininess(1) = (/60.0/)
    real, save :: back_mat_specular(4) = (/0.5, 0.5, 0.2, 1.0/)
    real, save :: back_mat_diffuse(4) = (/1.0, 1.0, 0.2, 1.0/)
    real, save :: lmodel_ambient(4) = (/1.0, 1.0, 1.0, 1.0/)
    real, save :: lmodel_twoside(1) = (/GL_TRUE/)

    call f90glMatrixMode(GL_PROJECTION)
    call f90gluPerspective(450.0_f90gldouble, 1.0_f90gldouble, 0.5_f90gldouble, 10.0_f90gldouble)

    call f90glLightfv(GL_LIGHT0, GL_AMBIENT, ambient)
    call f90glLightfv(GL_LIGHT0, GL_DIFFUSE, diffuse)
    call f90glLightfv(GL_LIGHT0, GL_POSITION, position)
    call f90glLightModelfv(GL_LIGHT_MODEL_AMBIENT, lmodel_ambient)
    call f90glLightModelfv(GL_LIGHT_MODEL_TWO_SIDE, lmodel_twoside)
    call f90glEnable(GL_LIGHTING)
    call f90glEnable(GL_LIGHT0)
    
    call f90glMaterialfv(GL_FRONT, GL_SHININESS, front_mat_shininess)
    call f90glMaterialfv(GL_FRONT, GL_SPECULAR, front_mat_specular)
    call f90glMaterialfv(GL_FRONT, GL_DIFFUSE, front_mat_diffuse)
    call f90glMaterialfv(GL_BACK, GL_SHININESS, back_mat_shininess)
    call f90glMaterialfv(GL_BACK, GL_SPECULAR, back_mat_specular)
    call f90glMaterialfv(GL_BACK, GL_DIFFUSE, back_mat_diffuse)
    if (rgb) then
        call f90glColorMaterial(GL_FRONT_AND_BACK, GL_DIFFUSE)
    endif

    if (rgb) then
        call f90glEnable(GL_COLOR_MATERIAL)
    else
        call SetColorMap()
    endif
end subroutine

subroutine InitTexture

    call f90glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT)
    call f90glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT)
    call f90glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST)
    call f90glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST)
    call f90glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE)
end subroutine

subroutine Init

    call f90glClearColor(0.0, 0.0, 0.0, 0.0)

    call f90glShadeModel(GL_FLAT)
    
    call f90glFrontFace(GL_CW)

    call f90glEnable(GL_DEPTH_TEST)

    call InitMaterials()
    call InitTexture()
    call InitMesh()

    call f90glMatrixMode(GL_MODELVIEW)
    call f90glTranslatef(0.0, 0.4, -1.8)
    call f90glScalef(2.0, 2.0, 2.0)
    call f90glRotatef(-35.0, 1.0, 0.0, 0.0)
    call f90glRotatef(35.0, 0.0, 0.0, 1.0)
end subroutine

end module

subroutine Reshape(width, height_loc)
use wave_mod
integer(f90glcint) width, height_loc
integer(f90glint) w,h

    w=width; h=height_loc
    call f90glViewport(0, 0, w, h)
end subroutine

function Key(k, mask)
use wave_mod
integer(f90glenum) Key
integer(f90glcint) k
integer(f90glenum) mask
external ftknull,animate

    real(f90GLfloat), save :: map(4) = (/0, 0, 20, 0/)

    Key = GL_TRUE
    select case (k) 
      case (TK_ESCAPE)
        call f90tkQuit()
      case (TK_lc_c)
        contouring = contouring + 1
        if (contouring == 1) then

            call f90glTexImage2D(GL_TEXTURE_2D, 0, 3, 4, 4, 0, GL_LUMINANCE, &
                         GL_UNSIGNED_BYTE, contourTexture1)
            call f90glTexGeni(GL_S, GL_TEXTURE_GEN_MODE, GL_OBJECT_LINEAR)
            call f90glTexGeni(GL_T, GL_TEXTURE_GEN_MODE, GL_OBJECT_LINEAR)
            call f90glTexGenfv(GL_S, GL_OBJECT_PLANE, map)
            call f90glTexGenfv(GL_T, GL_OBJECT_PLANE, map)
            call f90glEnable(GL_TEXTURE_2D)
            call f90glEnable(GL_TEXTURE_GEN_S)
            call f90glEnable(GL_TEXTURE_GEN_T)
        elseif (contouring == 2) then
            call f90glTexGeni(GL_S, GL_TEXTURE_GEN_MODE, GL_EYE_LINEAR)
            call f90glTexGeni(GL_T, GL_TEXTURE_GEN_MODE, GL_EYE_LINEAR)
            call f90glPushMatrix()
            call f90glMatrixMode(GL_MODELVIEW)
            call f90glLoadIdentity()
            call f90glTexGenfv(GL_S, GL_EYE_PLANE, map)
            call f90glTexGenfv(GL_T, GL_EYE_PLANE, map)
            call f90glPopMatrix()
        else
            contouring = 0
            call f90glDisable(GL_TEXTURE_GEN_S)
            call f90glDisable(GL_TEXTURE_GEN_T)
            call f90glDisable(GL_TEXTURE_2D)
        endif
      case (TK_lc_s)
        smooth = .not.smooth
        if (smooth) then
            call f90glShadeModel(GL_SMOOTH)
        else
            call f90glShadeModel(GL_FLAT)
        endif
      case (TK_lc_l)
        lighting = .not.lighting
        if (lighting) then
            call f90glEnable(GL_LIGHTING)
            call f90glEnable(GL_LIGHT0)
            if (rgb) &
                call f90glEnable(GL_COLOR_MATERIAL)
        else
            call f90glDisable(GL_LIGHTING)
            call f90glDisable(GL_LIGHT0)
            if (rgb) &
                call f90glDisable(GL_COLOR_MATERIAL)
        endif
      case (TK_lc_d)
        depth = .not.depth
        if (depth) then
            call f90glEnable(GL_DEPTH_TEST)
            clearMask = ior(clearMask,GL_DEPTH_BUFFER_BIT)
        else
            call f90glDisable(GL_DEPTH_TEST)
            clearMask = iand(clearMask,not(GL_DEPTH_BUFFER_BIT))
        endif
      case (TK_SPACE)
        stepMode = .not.stepMode
        if (stepMode) then
            call f90tkIdleFunc(ftknull)
            call f90tkDisplayFunc(Animate)
        else
            call f90tkIdleFunc(Animate)
            call f90tkDisplayFunc(ftknull)
        endif
      case (TK_lc_n)
        if (stepMode) then
            nextFrame = 1
        endif
      case (TK_lc_a)
        spinMode = .not.spinMode
      case default
        key = GL_FALSE
    end select
end function

subroutine Animate
use wave_mod

    type(coord_t), pointer :: coord
    type(facet_t), pointer :: facet
    real, pointer :: lastColor(:)
    real, pointer :: thisColor(:)
    integer(f90GLint) i, j
    logical temptest

    call f90glClear(clearMask)

    if (nextFrame/=0 .or. .not.stepMode) curFrame = curFrame + 1
    if (curFrame >= theMesh%frames) curFrame = 0

    if ((nextFrame/=0 .or. .not.stepMode) .and. spinMode) &
        call f90glRotatef(5.0, 0.0, 0.0, 1.0)

    nextFrame = 0

    do i = 0, theMesh%widthX-1
        call f90glBegin(GL_QUAD_STRIP)
        nullify(lastColor)
        do j = 0, theMesh%widthY-1
            facet => GETFACET(curFrame, i, j)
            if (.not.smooth .and. lighting) &
                call f90glNormal3fv(facet%normal)
            if (lighting) then
                if (rgb) then
                    thisColor => facet%color
                    call f90glColor3fv(facet%color)
                else
                    thisColor => facet%color
                    call f90glMaterialfv(GL_FRONT_AND_BACK, GL_COLOR_INDEXES, &
                                 facet%color)
                endif
            else
                if (rgb) then
                    thisColor => facet%color
                    call f90glColor3fv(facet%color)
                else
                    thisColor => facet%color
                    call f90glIndexf(facet%color(2))
                endif
            endif

            if (associated(lastcolor)) then
               temptest = thisColor(1) /= lastColor(1) .and. smooth
            else
               temptest = .true.
            endif
            if (.not.associated(lastColor) .or. temptest) then
                if (associated(lastColor)) then
                    call f90glEnd()
                    call f90glBegin(GL_QUAD_STRIP)
                endif
                coord => GETCOORD(curFrame, i, j)
                if (smooth .and. lighting) &
                    call f90glNormal3fv(coord%normal)
                call f90glVertex3fv(coord%vertex)

                coord => GETCOORD(curFrame, i+1, j)
                if (smooth .and. lighting) &
                    call f90glNormal3fv(coord%normal)
                call f90glVertex3fv(coord%vertex)
            endif

            coord => GETCOORD(curFrame, i, j+1)
            if (smooth .and. lighting) &
                call f90glNormal3fv(coord%normal)
            call f90glVertex3fv(coord%vertex)

            coord => GETCOORD(curFrame, i+1, j+1)
            if (smooth .and. lighting) &
                call f90glNormal3fv(coord%normal)
            call f90glVertex3fv(coord%vertex)

            lastColor => thisColor
        end do
        call f90glEnd()
    end do

    call f90glFlush()
    if (doubleBuffer) &
        call f90tkSwapBuffers()
end subroutine

program main
use wave_mod

    integer(f90GLenum) type
    external reshape,animate
    integer(f90glenum), external :: key

    rgb = .true.
    doubleBuffer = .true.
    directRender = .true.
    frames = 10
    widthX = 10
    widthY = 10
    checkerSize = 2
    height = 0.2

    call f90tkInitPosition(0, 0, 300, 300)

    type = TK_DEPTH
    if (rgb) then
       type = ior(type,TK_RGB)
    else
       type = ior(type,TK_INDEX)
    endif
    if (doubleBuffer) then
       type = ior(type,TK_DOUBLE)
    else
       type = ior(type,TK_SINGLE)
    endif
    if (directRender) then
       type = ior(type,TK_DIRECT)
    else
       type = ior(type,TK_INDIRECT)
    endif
    call f90tkInitDisplayMode(type)

    if (f90tkInitWindow("Wave Demo") == GL_FALSE) call f90tkQuit()

    call Init()

    call f90tkExposeFunc(Reshape)
    call f90tkReshapeFunc(Reshape)
    call f90tkKeyDownFunc(Key)
    call f90tkIdleFunc(Animate)
    call f90tkExec()
end program
